#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from datetime import datetime
from threading import local

# import interfaces

# import packages
from beaker import cache
from pyams_utils.lock import locked
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.view import view_config


_local = local()


PROGRESS_CACHE_NAME = 'PyAMS::progress'
PROGRESS_TASKS_CACHE_KEY = 'PyAMS::progress::running_tasks'
PROGRESS_LOCK_NAME = 'tasks_progress_lock'
PROGRESS_TASK_KEY = 'PyAMS::progress::task::{0}'


def get_tasks_cache():
    """Get cache storing tasks list"""
    try:
        tasks_cache = _local.running_tasks_cache
    except AttributeError:
        manager = cache.CacheManager(**cache.cache_regions['persistent'])
        tasks_cache = _local.running_tasks_cache = manager.get_cache(PROGRESS_CACHE_NAME)
    return tasks_cache


def get_progress_cache():
    """Get cache storing tasks progress"""
    try:
        local_cache = _local.progress_cache
    except AttributeError:
        manager = cache.CacheManager(**cache.cache_regions['default'])
        local_cache = _local.progress_cache = manager.get_cache(PROGRESS_CACHE_NAME)
    return local_cache


def get_running_tasks():
    """Get list of running tasks"""
    tasks_cache = get_tasks_cache()
    return tasks_cache.get_value(PROGRESS_TASKS_CACHE_KEY, createfunc=set)


def set_running_tasks(tasks):
    """Update list of running tasks"""
    tasks_cache = get_tasks_cache()
    tasks_cache.set_value(PROGRESS_TASKS_CACHE_KEY, tasks)


@locked(name=PROGRESS_LOCK_NAME)
def init_progress_status(progress_id, owner, label, tags=None, length=None, current=None):
    """Initialize progress status for given task

    :param str progress_id: task ID
    :param str owner: user ID associated with this task
    :param str label: label associated with this task
    :param tags: list of tags associated with given task
    :param int length: whole length of the given task, if available
    :param int current: current position in the whole task length, if available
    """
    status = {'status': 'running',
              'owner': owner,
              'label': label,
              'tags': tags,
              'length': length,
              'current': current,
              'started': datetime.utcnow()}
    # Store task status
    cache_key = PROGRESS_TASK_KEY.format(progress_id)
    progress_cache = get_progress_cache()
    progress_cache.set_value(cache_key, status)
    # Store task in running tasks list
    tasks = get_running_tasks()
    tasks.add(progress_id)
    set_running_tasks(tasks)


@locked(name=PROGRESS_LOCK_NAME)
def get_progress_status(progress_id):
    """Get status of given task"""
    progress_cache = get_progress_cache()
    cache_key = PROGRESS_TASK_KEY.format(progress_id)
    try:
        status = progress_cache.get_value(cache_key)
    except KeyError:
        status = {'status': 'unknown'}
    else:
        if status.get('status') == 'finished':
            progress_cache.remove_value(cache_key)
            tasks = get_running_tasks()
            if progress_id in tasks:
                tasks.remove(progress_id)
    return status


@locked(name=PROGRESS_LOCK_NAME)
def set_progress_status(progress_id, status='running', message=None, length=None, current=None):
    """Set status of given task

    :param str progress_id: task ID
    :param str status: new status of the given task
    :param str message: status message associated with the given task
    :param int length: whole length of the given task, if available
    :param int current: current position in the whole task length, if available
    """
    progress_cache = get_progress_cache()
    cache_key = PROGRESS_TASK_KEY.format(progress_id)
    try:
        task_status = progress_cache.get_value(cache_key)
    except KeyError:
        task_status = {'status': 'unknown'}
    task_status.update({'status': status,
                        'message': message,
                        'length': length,
                        'current': current})
    progress_cache.set_value(cache_key, task_status)


@view_config(name='get_progress_status.json', renderer='JSON', xhr=True)
def get_progress_status_view(request):
    """Get progress status of a given task

    Each submitted task is identified by an ID defined when the task is created
    """
    if 'progress_id' not in request.params:
        raise HTTPBadRequest("Missing argument")
    return get_progress_status(request.params['progress_id'])
